import { createHash } from '@s-ui/js/lib/hash';
import isNode from '../../helpers/isNode';
import isPromise from '../../helpers/isPromise';
import stringOrIntToMs from '../../helpers/stringOrIntToMs';
import LRU from './algorithms/LRU';
var ALGORITHMS = {
  LRU: 'lru'
};
var DEFAULT_TTL = 500;
var caches = {};

var _cache = function _cache(_temp) {
  var _ref = _temp === void 0 ? {} : _temp,
      algorithm = _ref.algorithm,
      fnName = _ref.fnName,
      instance = _ref.instance,
      original = _ref.original,
      size = _ref.size,
      target = _ref.target,
      ttl = _ref.ttl;

  caches[fnName] = caches[fnName] || (algorithm === ALGORITHMS.LRU ? new LRU({
    size: size
  }) : new Error("[sui-decorators::cache] unknown algorithm: " + algorithm));
  var cache = caches[fnName];
  return function () {
    for (var _len = arguments.length, args = new Array(_len), _key = 0; _key < _len; _key++) {
      args[_key] = arguments[_key];
    }

    if (typeof window !== 'undefined' && window.__SUI_CACHE_DISABLED__ || typeof global !== 'undefined' && global.__SUI_CACHE_DISABLED__) {
      return original.apply(instance, args);
    }

    var key = target.constructor.name + "::" + fnName + "::" + createHash(JSON.stringify(args));
    var now = Date.now();

    if (cache.get(key) === undefined) {
      cache.set(key, {
        createdAt: now,
        returns: original.apply(instance, args)
      });
    }

    if (isPromise(cache.get(key).returns)) {
      cache.get(key).returns.then(function (args) {
        if (args.__INLINE_ERROR__ && Array.isArray(args) && args[0] !== undefined) {
          cache.del(key);
        }
      })["catch"](function () {
        return cache.del(key);
      });
    }

    if (now - cache.get(key).createdAt > ttl) {
      cache.del(key);
    }

    return cache.get(key) !== undefined ? cache.get(key).returns : original.apply(instance, args);
  };
};

export default (function (_temp2) {
  var _ref2 = _temp2 === void 0 ? {} : _temp2,
      _ref2$ttl = _ref2.ttl,
      ttl = _ref2$ttl === void 0 ? DEFAULT_TTL : _ref2$ttl,
      _ref2$server = _ref2.server,
      server = _ref2$server === void 0 ? false : _ref2$server,
      _ref2$algorithm = _ref2.algorithm,
      algorithm = _ref2$algorithm === void 0 ? ALGORITHMS.LRU : _ref2$algorithm,
      size = _ref2.size;

  var timeToLife = stringOrIntToMs({
    ttl: ttl
  }) || DEFAULT_TTL;
  return function (target, fnName, descriptor) {
    // if we're on node but the decorator doesn't have the server flag
    // then we ignore the usage of the decorator and thus the cache
    if (isNode && !server) {
      return descriptor;
    }

    var configurable = descriptor.configurable,
        enumerable = descriptor.enumerable,
        writable = descriptor.writable;
    var originalGet = descriptor.get;
    var originalValue = descriptor.value;
    var isGetter = !!originalGet; // https://github.com/jayphelps/core-decorators.js/blob/master/src/autobind.js

    return Object.assign({}, {
      configurable: configurable,
      enumerable: enumerable,
      get: function get() {
        var fn = isGetter ? originalGet.call(this) : originalValue;

        if (this === target) {
          return fn;
        }

        var _fnCached = _cache({
          algorithm: algorithm,
          fnName: fnName,
          instance: this,
          original: fn,
          size: size,
          target: target,
          ttl: timeToLife
        });

        Object.defineProperty(this, fnName, {
          configurable: configurable,
          writable: writable,
          enumerable: enumerable,
          value: _fnCached
        });
        return _fnCached;
      },
      set: function set(newValue) {
        Object.defineProperty(this, fnName, {
          configurable: true,
          writable: true,
          enumerable: true,
          value: newValue
        });
        return newValue;
      }
    });
  };
});